import hashlib
from datetime import datetime
from typing import Any

import ecdsa
from uagents.crypto import Identity

from verifyaname.helpers import (
    AGENT_ADDRESS_PREFIX,
    SIGNATURE_PREFIX,
    decode_bech32,
    encode_length_prefixed
)

def _generate_digest(
    contract_address: str,
    domain: str,
    address: str,
    expiry: int,
    chain_id: str,
):
    hasher = hashlib.sha256()
    hasher.update(encode_length_prefixed(contract_address))
    hasher.update(encode_length_prefixed(domain))
    hasher.update(encode_length_prefixed(address))
    hasher.update(encode_length_prefixed(expiry))
    hasher.update(encode_length_prefixed(chain_id))

    return hasher.digest()


def _signed_digest(
    identity: Identity,
    contract_address: str,
    domain: str,
    address: str,
    expiry: int,
    chain_id: str,
) -> str:
    return identity.sign_digest(
        _generate_digest(contract_address, domain, address, expiry, chain_id)
    )


def generate_approval_token(
    identity: Identity,
    contract_address: str,
    domain: str,
    address: str,
    expiry: datetime,
    chain_id: str,
) -> Any:
    oracle_wallet_pubkey = identity.address
    seconds_since_epoch = int(expiry.timestamp())
    nanos_since_epoch = int(seconds_since_epoch * 10**9)

    signature = _signed_digest(
        identity, contract_address, domain, address, seconds_since_epoch, chain_id
    )

    verification_dict = {
        "oracle_agent_address": oracle_wallet_pubkey,
        "signature": signature,
        "expiry": str(nanos_since_epoch),
    }

    return verification_dict


def _verify_signature(
    signature: str,
    contract_address: str,
    domain: str,
    oracle_agent_address: str,
    address: str,
    expiry: datetime,
    chain_id: str,
) -> bool:
    # This is just a mock implementation of signature verification - it will be done onchain

    seconds_since_epoch = int(expiry.timestamp())
    digest = _generate_digest(
        contract_address, domain, address, seconds_since_epoch, chain_id
    )

    oracle_pk_prefix, oracle_pk_data = decode_bech32(oracle_agent_address)

    sig_prefix, sig_data = decode_bech32(signature)

    if oracle_pk_prefix != AGENT_ADDRESS_PREFIX:
        raise ValueError("Unable to decode oracle public key")

    if sig_prefix != SIGNATURE_PREFIX:
        raise ValueError("Unable to decode signature")

    # build the verifying key
    verifying_key = ecdsa.VerifyingKey.from_string(
        oracle_pk_data, curve=ecdsa.SECP256k1
    )

    try:
        result = verifying_key.verify_digest(sig_data, digest)
    except ecdsa.keys.BadSignatureError:
        return False

    return result


def verify_token(
    approval_token: dict,
    contract_address: str,
    domain: str,
    address: str,
    chain_id: str,
) -> bool:
    # This is just a mock implementation of approval token verification - it will be done onchain

    signature = approval_token["signature"]
    oracle_agent_address = approval_token["oracle_agent_address"]
    expiry = datetime.fromtimestamp(int(approval_token["expiry"]) // 10**9)

    # Onchain verification will consist of:
    # Verify if oracle_wallet_pubkey belongs to the oracle address
    # Verify if the address from owner_wallet_pubkey is the same as info.sender

    return _verify_signature(
        signature,
        contract_address,
        domain,
        oracle_agent_address,
        address,
        expiry,
        chain_id,
    )



import base64
import hashlib
import json
from dataclasses import dataclass

from cosmpy.aerial.wallet import Wallet
from cosmpy.crypto.address import Address
from cosmpy.crypto.keypairs import PublicKey
from verifyaname.helpers import encode_length_prefixed

TXT_VERIFICATION_KEY = "fetch-ans-token"
CHAIN_ID = "verification"


def _generate_digest(domain: str, address: str) -> bytes:
    hasher = hashlib.sha256()
    hasher.update(encode_length_prefixed(domain))
    hasher.update(encode_length_prefixed(address))
    return hasher.digest()


def generate_verification_string(wallet: Wallet, domain: str, address: str) -> str:
    signer = wallet.signer()
    signing_data = generate_signing_data(domain, address)

    signature = signer.sign(
        signing_data,
        deterministic=False,
        canonicalise=True,
    )

    return VerificationData(
        pubkey=wallet.public_key().public_key_bytes, signature=signature
    ).encode()


def pk_to_address(pk_bytes: bytes) -> str:
    return str(Address(PublicKey(pk_bytes), "fetch"))


def _signing_data_amino(
    from_address: str,
    data: bytes,
    chain_id: str = CHAIN_ID,
    account_number: int = 0,
    sequence: int = 0,
    memo: str = "",
):
    sign_doc_dict = {
        "account_number": str(account_number),
        "chain_id": str(chain_id),
        "fee": {"amount": [], "gas": "200000"},
        "memo": memo,
        "msgs": [
            {
                "type": "cosmos-sdk/MsgSignData",
                "value": {
                    "data": base64.b64encode(data).decode(),
                    "from_address": from_address,
                },
            }
        ],
        "sequence": str(sequence),
    }

    # Return normalized json string with sorted keys and no whitespace characters in binary format
    return json.dumps(
        sign_doc_dict, sort_keys=True, separators=(",", ":"), indent=None
    ).encode("utf-8")


def generate_signing_data(
    domain: str,
    address: str,
):
    digest = _generate_digest(domain, address)
    signing_data = _signing_data_amino(address, digest)

    return signing_data


@dataclass
class VerificationData:
    pubkey: bytes
    signature: bytes

    @staticmethod
    def _safe_base64_encode(data: bytes):
        """
        Removes any `=` used as padding from the encoded string.
        """
        encoded = base64.urlsafe_b64encode(data).decode()
        return encoded.rstrip("=")

    @staticmethod
    def _safe_base64_decode(data: str):
        """
        Adds back in the required padding before decoding.
        """
        padding = 4 - (len(data) % 4)
        string = data + ("=" * padding)
        return base64.urlsafe_b64decode(string)

    @staticmethod
    def decode(verification_string: str) -> "VerificationData":
        decoded_verification = VerificationData._safe_base64_decode(verification_string)
        pk_bytes = decoded_verification[:33]
        signature = decoded_verification[33:]
        return VerificationData(pk_bytes, signature)

    def encode(self) -> str:
        return VerificationData._safe_base64_encode(self.pubkey + self.signature)


def verify_domain_string(
    data: VerificationData,
    domain: str,
    address: str,
) -> bool:
    # Verify the signature
    signing_data = generate_signing_data(domain, address)

    is_verified = PublicKey(data.pubkey).verify(signing_data, data.signature)

    return is_verified
